#!/usr/bin/env python

"""
Compliance Checker
"""
import csv
import itertools
import pprint
import re
import warnings
from collections import defaultdict
from functools import wraps
from io import StringIO

import validators
from lxml import etree
from netCDF4 import Dataset
from owslib.namespaces import Namespaces
from owslib.swe.observation.sos100 import SensorObservationService_1_0_0
from owslib.swe.sensor.sml import SensorML

from compliance_checker import __version__
from compliance_checker.util import kvp_convert

# Python 3.5+ should work, also have a fallback
try:
    from re import Pattern

    re_pattern_type = Pattern
except ImportError:
    re_pattern_type = type(re.compile(""))


def get_namespaces():
    n = Namespaces()
    ns = n.get_namespaces(["ogc", "sml", "gml", "sos", "swe", "xlink"])
    ns["ows"] = n.get_namespace("ows110")
    return ns


def csv_splitter(input_string):
    """
    csv_splitter(input_string)

    Splits a string in CSV format and returns a flattened list

    Parameters:
    -----------
    input_string: str
        The string to be processed

    Returns:
    --------
    list of str
        A flattened list from the CSV processing contents
    """
    csv_contents = csv.reader(StringIO(input_string))
    return list(itertools.chain.from_iterable(csv_contents))


class ValidationObject:
    validator_fail_msg = ""
    expected_type = None

    def __init__(self, split_func=None):
        if split_func is None:
            self.split_func = lambda x: [x]
        else:
            self.split_func = split_func

    def validator_func(self, input_value):
        """
        validator_func(self, input_value)

        Function that should validate the result of a given input value
        """
        raise NotImplementedError

    def validate(self, input_name, input_value):
        if self.expected_type is not None:
            type_result = self.validate_type(input_name, input_value)
            if not type_result[0]:
                return type_result
        for processed_value in self.split_func(input_value):
            validator_result = self.validator_func(processed_value)
            if not validator_result:
                return False, [self.validator_fail_msg.format(input_name)]
        # if all pass, then we're good.
        return True, None

    def validate_type(self, input_name, input_value):
        if not isinstance(input_value, self.expected_type):
            expected_type_fmt = "Attribute {} should be instance of type {}"
            return (
                False,
                [expected_type_fmt.format(input_name, self.expected_type.__name__)],
            )
        else:
            return True, None


class EmailValidator(ValidationObject):
    validator_fail_msg = "{} must be a valid email address"
    expected_type = str

    def validator_func(self, input_value):
        return validators.email(input_value)


class RegexValidator(ValidationObject):
    expected_type = str
    validator_regex = r"^.+$"
    validator_fail_msg = "{} must not be an empty string"

    def validator_func(self, input_value):
        return bool(re.search(self.validator_regex, input_value))


class UrlValidator(ValidationObject):
    validator_fail_msg = "{} must be a valid URL"
    expected_type = str

    def validator_func(self, input_value):
        return bool(validators.url(input_value))


# Simple class for Generic File type (default to this if file not recognised)
class GenericFile:
    """
    Simple class for any file. Has same path lookup as netCDF4.Dataset.
    """

    def __init__(self, fpath):
        self.fpath = fpath

    def filepath(self):
        return self.fpath


class BaseCheck:
    HIGH = 3
    MEDIUM = 2
    LOW = 1

    _cc_checker_version = __version__
    _cc_display_headers = {3: "High Priority", 2: "Medium Priority", 1: "Low Priority"}

    supported_ds = []

    def setup(self, ds):
        """
        Common setup method for a Checker.

        Automatically run when running a CheckSuite. Define this method in your Checker class.
        """

    def __init__(self, options=None):
        self._defined_results = defaultdict(lambda: defaultdict(dict))
        if options is None:
            self.options = {}
        else:
            self.options = options

    def get_test_ctx(self, severity, name, variable=None):
        """
        Creates an existing TestCtx object in _defined_results dict if it does
        not exist for the current checker instance, or an returns the existing
        TestCtx for modification. Takes a severity level and name and uses the
        two element tuple formed by the arguments as a key into the dict.

        :param int severity: A BaseCheck severity level
        :param str name: The name of the check
        :rtype compliance_checker.base.TestCtx:
        :returns: A new or or existing `TestCtx` instance taken from this
                  instance's _defined_results dict
        """
        # Is it necessary to key out by severity?  Is severity level unique
        # per check?  If so, it could be eliminated from key hierarchy
        if severity not in self._defined_results[name][variable]:
            self._defined_results[name][variable][severity] = TestCtx(
                severity,
                name,
                variable=variable,
            )
        return self._defined_results[name][variable][severity]


class BaseNCCheck:
    """
    Base Class for NetCDF Dataset supporting Check Suites.
    """

    supported_ds = [Dataset]

    @classmethod
    def std_check_in(cls, dataset, name, allowed_vals):
        """
        Returns 0 if attr not present, 1 if present but not in correct value, 2 if good
        """
        if name not in dataset.ncattrs():
            return 0

        ret_val = 1
        if dataset.getncattr(name) in allowed_vals:
            ret_val += 1

        return ret_val

    @classmethod
    def std_check(cls, dataset, name):
        return name in dataset.ncattrs()


class BaseSOSGCCheck:
    """
    Base class for SOS-GetCapabilities supporting Check Suites.
    """

    supported_ds = [SensorObservationService_1_0_0]


class BaseSOSDSCheck:
    """
    Base class for SOS-DescribeSensor supporting Check Suites.
    """

    supported_ds = [SensorML]


class Result:
    """
    Holds the result of a check method.

    Stores such information as the check's value (True, False, a 2-tuple of (pass, total) or None for a skip),
    weight of the check, any granular messages, or a hierarchy of results. If given value is not a tuple, it
    is cast as a boolean using the bool() function.

    Stores the checker instance and the check method that produced this result.
    """

    def __init__(
        self,
        weight=BaseCheck.MEDIUM,
        value=None,
        name=None,
        msgs=None,
        children=None,
        checker=None,
        check_method=None,
        variable_name=None,
    ):
        self.weight = weight

        if value is None:
            self.value = None
        elif isinstance(value, tuple):
            if len(value) != 2:
                raise ValueError(
                    f"Result value must be 2-tuple or boolean! Got {value}",
                )
            self.value = value
        else:
            self.value = bool(value)
        self.name = name
        self.msgs = msgs or []

        self.children = children or []

        self.checker = checker
        self.check_method = check_method
        self.variable_name = variable_name

    def __repr__(self):
        ret = f"{self.name} (*{self.weight}): {self.value}"

        if len(self.msgs):
            if len(self.msgs) == 1:
                ret += f" ({self.msgs[0]})"
            else:
                ret += f" ({len(self.msgs)!s} msgs)"

        if len(self.children):
            ret += f" ({len(self.children)!s} children)"
            ret += "\n" + pprint.pformat(self.children)

        return ret

    def serialize(self):
        """
        Returns a serializable dictionary that represents the result object
        """
        return {
            "name": self.name,
            "weight": self.weight,
            "value": self.value,
            "msgs": self.msgs,
            "children": [i.serialize() for i in self.children],
        }

    def __eq__(self, other):
        return self.serialize() == other.serialize()


class TestCtx:
    """
    Simple struct object that holds score values and messages to compile into a result
    """

    def __init__(
        self,
        category=None,
        description="",
        out_of=0,
        score=0,
        messages=None,
        variable=None,
    ):
        self.category = category or BaseCheck.LOW
        self.out_of = out_of
        self.score = score
        self.messages = messages or []
        self.description = description or ""
        self.variable = variable

    def to_result(self):
        return Result(
            self.category,
            (self.score, self.out_of),
            self.description,
            self.messages,
            variable_name=self.variable,
        )

    def assert_true(self, test, message):
        """
        Increments score if test is true otherwise appends a message
        :rtype: bool
        :return: Boolean indicating whether test condition passed or not
        """
        self.out_of += 1

        if test:
            self.score += 1
        else:
            self.messages.append(message)

        return test

    def add_failure(self, message):
        """
        Adds a failure along with a message
        :rtype: None
        """
        self.assert_true(False, message)

    def add_pass(self):
        """
        Adds a pass condition
        :rtype: None
        """
        self.assert_true(True, None)


def std_check_in(base_context, name, allowed_vals):
    """
    Check that a value is contained within an iterable

    Parameters:
    -----------
    base_context: netCDF4.Dataset or netCDF4.variable
       The context in which to look for the attribute, either a
       netCDF4.Dataset or netCDF4.Variable.  If a netCDF dataset,
       the attribute is searched for in the global attributes.
       If a variable, the attributes are limited to those contained
       in the corresponding variable.
    name: str
       The name of the attribute to search for.
    allowed_vals: iterable
       An iterable, usually a set, which provides the possible valid values for
       the attribute.
    Returns:
    --------
    int
        Returns 0 if attr not present, 1 if present but not in correct value, 2
        if good.
    """
    if not hasattr(base_context, name):
        return 0

    ret_val = 1
    if base_context.getncattr(name) in allowed_vals:
        ret_val += 1

    return ret_val


def std_check(dataset, name):
    if hasattr(dataset, name):
        getattr(dataset, name)
        return True

    return False


def xpath_check(tree, xpath):
    """Checks whether tree contains one or more elements matching xpath"""
    return len(xpath(tree)) > 0


def maybe_get_global_attr(attr_name, ds):
    if attr_name in ds.ncattrs():
        return True, ds.getncattr(attr_name)
    else:
        err_msg = "{} not present"
        return False, [err_msg.format(attr_name)]


def attr_check(kvp, ds, priority, ret_val, gname=None, var_name=None):
    """
    Handles attribute checks for simple presence of an attribute, presence of
    one of several attributes, and passing a validation function.  Returns a
    status along with an error message in the event of a failure.  Mutates
    ret_val parameter

    :param tuple(str, func) or str l: the attribute being checked
    :param netCDF4 dataset ds       : dataset being checked
    :param int priority             : priority level of check
    :param list ret_val             : result to be returned
    :param str or None gname        : group name assigned to a group of attribute Results
    :param str or None var_name     : name of the variable which contains this attribute
    """

    msgs = []
    name, other = kvp
    if var_name is not None:
        display_name = f"attribute {name} in variable {var_name}"
        base_context = ds.variables[var_name]
    else:
        display_name = name
        base_context = ds
    if other is None:
        res = std_check(ds, name)
        if not res:
            msgs = [f"{display_name} not present"]
        else:
            try:
                # see if this attribute is a string, try stripping
                # whitespace, and return an error if empty
                att_strip = base_context.getncattr(name).strip()
                if not att_strip:
                    res = False
                    msgs = [f"{display_name} is empty or completely whitespace"]
            # if not a string/has no strip method we should be OK
            except AttributeError:
                pass

        # gname arg allows the global attrs to be grouped together
        ret_val.append(
            Result(
                priority,
                value=res,
                name=gname if gname else name,
                msgs=msgs,
                variable_name=var_name,
            ),
        )
    elif hasattr(other, "__iter__"):
        # redundant, we could easily do this with a hasattr
        # check instead
        res = std_check_in(base_context, name, other)
        if res == 0:
            msgs.append(f"{display_name} not present")
        elif res == 1:
            msgs.append(
                f"{display_name} present, but not in expected value list ({sorted(other)})",
            )

        ret_val.append(
            Result(
                priority,
                (res, 2),
                gname if gname else name,  # groups Globals if supplied
                msgs,
                variable_name=var_name,
            ),
        )
    # if we have an XPath expression, call it on the document
    elif type(other) is etree.XPath:
        # TODO: store tree instead of creating it each time?
        # no execution path for variable
        res = xpath_check(ds._root, other)
        if not res:
            msgs = [f"XPath for {display_name} not found"]
        ret_val.append(
            Result(
                priority,
                res,
                gname if gname else name,
                msgs,
                variable_name=var_name,
            ),
        )
    # check if this is a subclass of ValidationObject
    elif isinstance(other, ValidationObject):
        attr_result = maybe_get_global_attr(name, ds)
        if not attr_result[0]:
            res_tup = attr_result
        else:
            check_val = attr_result[1]
            res_tup = other.validate(name, check_val)

        msgs = [] if res_tup[1] is None else res_tup[1]

        ret_val.append(Result(priority, res_tup[0], name, msgs))
    elif isinstance(other, re_pattern_type):
        attr_result = maybe_get_global_attr(name, ds)
        if not attr_result[0]:
            return attr_result
        else:
            check_val = attr_result[1]
        if not isinstance(check_val, str):
            res = False
            msgs = [f"{name} must be a string"]
        elif not other.search(check_val):
            res = False
            msgs = [f"{name} must match regular expression {other}"]
        else:
            res = True
            msgs = []

        ret_val.append(
            Result(priority, value=res, name=gname if gname else name, msgs=msgs),
        )

    # if the attribute is a function, call it
    # right now only supports single attribute
    # important note: current magic approach uses all functions
    # starting with "check".  Avoid naming check functions
    # starting with check if you want to pass them in with
    # a tuple to avoid them being checked more than once
    elif callable(other):
        # check that the attribute is actually present.
        # This reduces boilerplate in functions by not needing
        # to check whether the attribute is present every time
        # and instead focuses on the core functionality of the
        # test

        res = other(base_context)  # call the method on the dataset
        if not res:
            msgs = [f"{display_name} not present"]
            ret_val.append(
                Result(
                    priority,
                    res,
                    gname if gname else name,
                    msgs,
                    variable_name=var_name,
                ),
            )
        else:
            ret_val.append(res(priority))
    # unsupported second type in second
    else:
        raise TypeError(
            f"Second arg in tuple has unsupported type: {type(other)}",
        )

    return ret_val


def check_has(priority=BaseCheck.HIGH, gname=None):
    """Decorator to wrap a function to check if a dataset has given attributes.
    :param function func: function to wrap"""

    def _inner(func):
        def _dec(s, ds):
            attr_process = kvp_convert(func(s, ds))

            ret_val = []
            # could potentially run tests in parallel if we eliminated side
            # effects on `ret_val`
            for kvp in attr_process.items():
                # function mutates ret_val
                attr_check(kvp, ds, priority, ret_val, gname)
            return ret_val

        return wraps(func)(_dec)

    return _inner


def fix_return_value(v, method_name, method=None, checker=None):
    """
    Transforms scalar return values into Result.
    """
    # remove common check prefix
    method_name = (method_name or method.__func__.__name__).replace("check_", "")
    if v is None or not isinstance(v, Result):
        v = Result(value=v, name=method_name)

    v.name = v.name or method_name

    v.checker = checker
    v.check_method = method

    return v


def ratable_result(value, name, msgs, variable_name=None):
    """Returns a partial function with a Result that has not been weighted."""
    return lambda w: Result(w, value, name, msgs, variable_name=variable_name)


def score_group(group_name=None):
    """
    Warning this is deprecated as of Compliance Checker v3.2!

    Please do not using scoring groups and update your plugins
    if necessary
    """
    warnings.warn(
        "Score_group is deprecated as of Compliance Checker v3.2.",
        stacklevel=2,
    )

    def _inner(func):
        def _dec(s, ds):
            ret_val = func(s, ds)
            """
            if group_name != None and not isinstance(ret_val[0], tuple):
                return tuple([(group_name, ret_val[0])] + list(ret_val[1:]))
            """
            # multiple returns
            if not isinstance(ret_val, list):
                ret_val = [ret_val]

            def dogroup(r):
                cur_grouping = r.name

                if isinstance(cur_grouping, tuple):
                    cur_grouping = list(cur_grouping)
                elif not isinstance(cur_grouping, list):
                    cur_grouping = [cur_grouping]

                cur_grouping.insert(0, group_name)

                return Result(r.weight, r.value, tuple(cur_grouping), r.msgs)

            ret_val = [fix_return_value(x, func.__name__, func, s) for x in ret_val]
            ret_val = list(map(dogroup, ret_val))

            return ret_val

        return wraps(func)(_dec)

    return _inner



#!/usr/bin/env python
"""
Tests for command line output and parsing

"""
import importlib.util
import io
import json
import os
import platform
import subprocess
import sys
from argparse import Namespace
from collections import defaultdict
from importlib.machinery import SourceFileLoader

import pytest

from compliance_checker.runner import CheckSuite, ComplianceChecker

from .conftest import datadir, static_files

on_windows = platform.system() == "Windows"

if on_windows:
    ncconfig = ["sh", f"{os.environ['CONDA_PREFIX']}\\Library\\bin\\nc-config"]
else:
    ncconfig = ["nc-config"]


@pytest.mark.usefixtures("checksuite_setup")
class TestCLI:
    """
    Tests various functions and aspects of the command line tool and runner
    """

    def shortDescription(self):
        return None

    def test_unicode_acdd_html(self, tmp_txt_file):
        """
        Tests that the checker is capable of producing HTML with unicode characters
        """
        return_value, errors = ComplianceChecker.run_checker(
            ds_loc=static_files("2dim"),
            verbose=0,
            criteria="strict",
            checker_names=["acdd"],
            output_filename=tmp_txt_file,
            output_format="html",
        )

        assert os.stat(tmp_txt_file).st_size > 0

    def test_unicode_cf_html(self, tmp_txt_file):
        """
        Tests that the CF checker can produce HTML output with unicode characters
        """
        return_value, errors = ComplianceChecker.run_checker(
            ds_loc=static_files("2dim"),
            verbose=0,
            criteria="strict",
            checker_names=["cf"],
            output_filename=tmp_txt_file,
            output_format="html",
        )

        assert os.stat(tmp_txt_file).st_size > 0

    def test_single_json_output(self, tmp_txt_file):
        """
        Tests that a suite can produce JSON output to a file
        """
        return_value, errors = ComplianceChecker.run_checker(
            ds_loc=static_files("conv_bad"),
            verbose=0,
            criteria="strict",
            checker_names=["cf"],
            output_filename=tmp_txt_file,
            output_format="json",
        )

        assert os.stat(tmp_txt_file).st_size > 0
        with open(tmp_txt_file) as f:
            r = json.load(f)
            assert "cf" in r

    def test_list_checks(self):
        """
        Tests listing of both old-style, deprecated checkers using .name
        attributes, and newer ones which use ._cc_spec and _cc_spec_version
        attributes
        """

        # hack: use argparse.Namespace to mock checker object with attributes
        # since SimpleNamespace is Python 3.3+ only
        CheckSuite.checkers.clear()
        # need to mock setuptools entrypoints here in order to load in checkers

        def checker_1():
            return Namespace(name="checker_1")

        def checker_2():
            return Namespace(_cc_spec="checker_2", _cc_spec_version="2.2")

        mock_checkers = [Namespace(load=checker_1), Namespace(load=checker_2)]
        with pytest.warns(DeprecationWarning):
            CheckSuite._load_checkers(mock_checkers)

        cs = CheckSuite()
        saved = sys.stdout
        try:
            # ugly!  consider refactoring to use py.test capsys
            fake_stdout = io.StringIO()
            sys.stdout = fake_stdout
            cs._print_suites()
            assert fake_stdout.getvalue() == " - checker_1:unknown\n - checker_2:2.2\n"
        finally:
            sys.stdout = saved
            fake_stdout.close()

    def test_multiple_json_output(self, tmp_txt_file):
        """
        Tests that a suite can produce JSON output to a file
        """
        return_value, errors = ComplianceChecker.run_checker(
            ds_loc=static_files("conv_bad"),
            verbose=0,
            criteria="strict",
            checker_names=["acdd", "cf"],
            output_filename=tmp_txt_file,
            output_format="json",
        )

        assert os.stat(tmp_txt_file).st_size > 0
        with open(tmp_txt_file) as f:
            r = json.load(f)
            assert "cf" in r
            assert "acdd" in r

    def test_multiple_json_output_stdout(self):
        """
        Tests that a suite can produce JSON output to stdout
        """
        saved = sys.stdout
        try:
            fake_stdout = io.StringIO()
            sys.stdout = fake_stdout
            return_value, errors = ComplianceChecker.run_checker(
                ds_loc=static_files("conv_bad"),
                verbose=0,
                criteria="strict",
                checker_names=["acdd", "cf"],
                output_filename="-",
                output_format="json",
            )
            r = json.loads(fake_stdout.getvalue().strip())
            assert "acdd" in r
            assert "cf" in r
        finally:
            sys.stdout = saved
            fake_stdout.close()

    def test_single_json_output_stdout(self):
        """
        Tests that a suite can produce JSON output to stdout
        """
        saved = sys.stdout
        try:
            fake_stdout = io.StringIO()
            sys.stdout = fake_stdout
            return_value, errors = ComplianceChecker.run_checker(
                ds_loc=static_files("conv_bad"),
                verbose=0,
                criteria="strict",
                checker_names=["cf"],
                output_filename="-",
                output_format="json",
            )
            r = json.loads(fake_stdout.getvalue().strip())
            assert "cf" in r
        finally:
            sys.stdout = saved
            fake_stdout.close()

    def test_text_output(self, tmp_txt_file):
        """
        Tests that the 'text' output can be redirected to file with arguments
        to the command line
        """
        return_value, errors = ComplianceChecker.run_checker(
            ds_loc=static_files("conv_bad"),
            verbose=0,
            criteria="strict",
            checker_names=["acdd", "cf"],
            output_filename=tmp_txt_file,
            output_format="text",
        )

        assert os.stat(tmp_txt_file).st_size > 0

    def test_multi_checker_return_value(self, tmp_txt_file):
        """
        Tests that any failure for multiple checkers results in a failure return
        status
        """
        # CF should pass here
        return_value, errors = ComplianceChecker.run_checker(
            ds_loc=static_files("ncei_gold_point_1"),
            verbose=0,
            criteria="strict",
            checker_names=["cf:1.6"],
            output_filename=tmp_txt_file,
            output_format="text",
        )
        assert return_value

        # CF should pass, but ACDD will have some errors.  Overall return
        # status should be a failure
        return_value, errors = ComplianceChecker.run_checker(
            ds_loc=static_files("ncei_gold_point_1"),
            verbose=0,
            criteria="strict",
            checker_names=["acdd", "cf"],
            output_filename=tmp_txt_file,
            output_format="text",
        )
        assert not return_value

    def _check_libnetcdf_version():
        return (
            float(
                subprocess.check_output(
                    ncconfig + ["--version"],
                    encoding="UTF-8",
                )[9:12],
            )
            < 8.0
        )

    @pytest.mark.skipif(
        subprocess.check_output(ncconfig + ["--has-nczarr"]) != b"yes\n",
        reason="NCZarr is not available.",
    )
    @pytest.mark.parametrize(
        "zarr_url",
        [
            f"{(datadir / 'trajectory.zarr').as_uri()}#mode=nczarr,file",
            str(datadir / "trajectory.zarr.zip"),
            # "s3://hrrrzarr/sfc/20210408/20210408_10z_anl.zarr#mode=nczarr,s3"
        ],
        ids=[
            "local_file",
            "zip_file",
            # TODO uncomment once S3 support is working.
            # "s3_url",
        ],
    )
    def test_nczarr_pass_through(self, zarr_url):
        """
        Test that the url's with #mode=nczarr option pass through to ncgen
        https://www.unidata.ucar.edu/blogs/developer/entry/overview-of-zarr-support-in
        """
        skip_checks = [
            "check_filename",  # .zarr cannot pass a test that requires it to be named .nc
            "check_coordinate_variables_strict_monotonicity",  # FIXME: I believe there is a real problem with the original test data!
        ]
        return_value, errors = ComplianceChecker.run_checker(
            skip_checks=skip_checks,
            ds_loc=zarr_url,
            verbose=0,
            criteria="strict",
            checker_names=["cf:1.6"],
            output_format="text",
        )
        assert not errors


def test_parse_options():
    """Test the option parser of cchecker.py"""
    # Load cchecker.py
    cchecker_file_path = os.path.join(
        os.path.dirname(__file__),
        "..",
        "..",
        "cchecker.py",
    )
    spec = importlib.util.spec_from_file_location("cchecker", cchecker_file_path)
    module = importlib.util.module_from_spec(spec)
    SourceFileLoader(spec.name, spec.origin).exec_module(module)
    # Simple test checker_type:checker_opt
    opt_dict = module.parse_options(["cf:enable_appendix_a_checks"])
    assert opt_dict == defaultdict(dict, {"cf": {"enable_appendix_a_checks": None}})
    # Test case checker_type:checker_opt:checker_val
    opt_dict = module.parse_options(
        ["type:opt:val", "type:opt2:val:2", "cf:enable_appendix_a_checks"],
    )
    assert opt_dict == defaultdict(
        dict,
        {
            "type": {"opt": "val", "opt2": "val:2"},
            "cf": {"enable_appendix_a_checks": None},
        },
    )



try:
    import cf_units
except ImportError:
    cf_units = False

UT_SYSTEM = None

if cf_units:
    PyUdunits2 = cf_units.Unit
else:
    import pyudunits2

    class PyUdunits2:
        """Workaround for the differences in pyudunits2 and cf-units.

        NB: Some of these may change and/or get implemented upstream. Pyudunits2 is new and in-flux.

        1/4 Raise the same ValueError to match cf-unit errors.
        2/4 Creates an empty unit from None to mimic cf-unit's Unit('unknown')
        3/4 Add a definition object that is ust units.expanded()
        """

        def __init__(self, units: str | None):
            """Keep unit system so we can convert from string later."""
            global UT_SYSTEM
            if UT_SYSTEM is None:
                UT_SYSTEM = pyudunits2.UnitSystem.from_udunits2_xml()

            self.ut_system = UT_SYSTEM

            if units is None:
                units = ""

            try:
                self.units = self.ut_system.unit(units)
            except (SyntaxError, pyudunits2.UnresolvableUnitException) as err:
                raise ValueError from err
            self.definition = self.units.expanded()

        def __eq__(self, other):
            return self.units == other

        def is_dimensionless(self):
            return self.units.is_dimensionless()

        def is_time_reference(self):
            return isinstance(self.units, pyudunits2.DateUnit)

        def is_convertible(self, other):
            if isinstance(other, str):
                other = self.ut_system.unit(other)
            elif isinstance(other, (PyUdunits2)):
                other = other.units
            else:
                msg = f"Expected valid unit string or pyudunits2 unit object. Got {other}."
                raise ValueError(msg)

            # FIXME: cf-units Workaround 1/4 -> cf_units.Unit(None) -> Unit('unknown').
            if "" in (self.units.expanded(), other.expanded()):
                return False

            convertible = self.units.is_convertible_to(other)
            # FIXME: cf-units Workaround 2/4 -> time is not convertible to time reference.

            # Both are time reference confirm.
            if self.is_time_reference() and isinstance(other, pyudunits2.DateUnit):
                convertible = True

            return convertible


class Unit(PyUdunits2):
    def __init__(self, units):
        super().__init__(units)



"""Appendix C - Standard Name modifiers"""

# Dict of standard name modifiers with values for units.  "u" indicates to use
# the same units as the standard name canonical units, "1" is unitless for
# observation counts, and None is used for status_flag, which expects units
# not to be present
valid_modifiers = {
    "detection_minimum": "u",
    "number_of_observations": "1",
    "standard_error": "u",
    "status_flag": None,
}



#!/usr/bin/env python
"""
compliance_checker/tests/test_protocols.py

Unit tests that ensure the compliance checker can successfully identify protocol endpoints
"""
import platform

import pytest

from compliance_checker.protocols import zarr
from compliance_checker.suite import CheckSuite

from .conftest import datadir

pytestmark = [pytest.mark.integration]


@pytest.fixture
def cs():
    return CheckSuite()


@pytest.mark.vcr()
def test_netcdf_content_type(cs):
    """
    Check that urls with Content-Type header of "application/x-netcdf" can
    successfully be read into memory for checks.
    """
    url = "https://gliders.ioos.us/erddap/tabledap/amelia-20180501T0000.ncCF?&time%3E=max(time)-1%20hour"
    ds = cs.load_dataset(url)
    assert ds is not None


def test_erddap(cs):
    """
    Tests that a connection can be made to ERDDAP's GridDAP
    """
    url = "https://www.neracoos.org/erddap/griddap/WW3_EastCoast_latest"
    ds = cs.load_dataset(url)
    assert ds is not None


def test_hyrax(cs):
    """
    Tests that a connection can be made to Hyrax

    NOTE: This test creates a VCR file that is larger than GH's limit.
    We are not recording.
    """
    # Returns: error 405
    # url = "http://test.opendap.org:8080/opendap/ioos/mday_joinExist.ncml"
    # More direct file
    url = "http://test.opendap.org:8080/opendap/ioos/mday_joinExist.ncml.dap.nc4"
    ds = cs.load_dataset(url)
    assert ds is not None


def test_thredds(cs):
    """
    Tests that a connection can be made to a remote THREDDS endpoint

    NOTE: This test creates a VCR file that is larger than GH's limit.
    We are not recording.
    """
    # Returns: error 400
    # url = "http://thredds.ucar.edu/thredds/dodsC/grib/NCEP/GFS/Global_0p25deg_ana/TP"
    # Use a smaller dataset
    url = "https://thredds.ucar.edu/thredds/ncss/grid/grib/NCEP/GFS/Global_0p25deg_ana/TP?var=Temperature_altitude_above_msl&accept=netcdf3"

    cs = CheckSuite()
    ds = cs.load_dataset(url)
    assert ds is not None


str_dir = str(datadir.resolve()).replace("\\", "/")
file_url = "file://" + str_dir + "/trajectory.zarr#mode=nczarr,file"
s3_url = "s3://hrrrzarr/sfc/20210408/20210408_10z_anl.zarr#mode=nczarr,s3"
zip_url = "file://" + str_dir + "/trajectory.zarr.zip#mode=nczarr,zip"
# replace slashes for windows compatibility
url_io = [
    ("s3://hrrrzarr/sfc/20210408/20210408_10z_anl.zarr", s3_url),
    (s3_url, s3_url),
    (datadir / "trajectory.zarr", file_url),
    ("file://" + str_dir + "/trajectory.zarr", file_url),
    (file_url, file_url),
    (datadir / "trajectory.zarr.zip", zip_url),
    ("file://" + str_dir + "/trajectory.zarr.zip", zip_url),
    (zip_url, zip_url),
]


@pytest.mark.skipif(
    platform.system() != "Linux",
    reason="NCZarr is not officially supported for your OS as of when this API was written",
)
@pytest.mark.parametrize("url_in, url_out", url_io)
def test_as_zarr(url_in, url_out):
    """
    Test is as_zurl can transform pointers to zarr datasets to valid nczarr urls,
    """
    assert zarr.as_zarr(url_in) == url_out



#!/usr/bin/env python3
#
# compliance_checker documentation build configuration file, created by
# sphinx-quickstart on Mon Oct  9 21:28:42 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

from datetime import datetime

import compliance_checker

version = release = compliance_checker.__version__
cur_date = datetime.utcnow()


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
needs_sphinx = "2.1"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "myst_parser",
    "numpydoc",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
]

# Make MyST generate anchors for headings
myst_heading_anchors = 2

autodoc_typehints = "description"

# Tweak how docs are formatted
napoleon_use_rtype = False

# Control main class documentation
autoclass_content = "both"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# Custom sidebar templates, maps document names to template names.
html_sidebars = {"**": ["versions", "sidebar-nav-bs"]}

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
source_suffix = [".rst", ".md"]

# Controlling automatically generating summary tables in the docs
autosummary_generate = True
autosummary_ignore_module_all = False

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "compliance_checker"
copyright = f"{cur_date:%Y}, Dave Foster"
author = "Dave Foster"


# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    "external_links": [
        {
            "name": "Release Notes",
            "url": "https://github.com/ioos/compliance-checker/releases",
        },
    ],
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/ioos/compliance-checker",
            "icon": "fab fa-github-square",
        },
    ],
    "use_edit_page_button": False,
    "navbar_align": "left",
    "navbar_start": ["navbar-logo"],
    "navbar_center": ["navbar-nav"],
    "navbar_end": ["search-field", "navbar-icon-links"],
}


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "compliance_checkerdoc"


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, "compliance_checker", "compliance_checker Documentation", [author], 1),
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "compliance_checker",
        "compliance_checker Documentation",
        author,
        "compliance_checker",
        "One line description of project.",
        "Miscellaneous",
    ),
]

# GitHub issuecomment anchors are returning as broken for some reason.
linkcheck_ignore = ["https://github.com/ioos/compliance-checker/issues/.*"]
